/*
 * Copyright (C) 2022 luca0N!
 *
 * This file is part of Static Website Generator (swg).
 *
 * Static Website Generator (swg) is free software: you can redistribute it
 * and/or modify it under the terms of the version 3 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * Static Website Generator (swg) is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Static Website Generator (swg). If not, see
 * <https://www.gnu.org/licenses/>.
 *
 * Contact luca0N! by e-mail: <luca0n [at] luca0n [dot] com>. 
 */

#include <iostream>
#include <cstring>
#include <string>
#include <list>
#include <regex>

#include <filesystem>

#include "SwgRuntime.hxx"
#include "SwgContext.hxx"
#include "ConfigUtils.hxx"
// #include "Blog.h"
#include "Common.hxx"
#include "MarkdownParser.hxx"

std::string blog_relative_path(std::string const &pathPrefix, std::string const &path) {
	return path.substr(pathPrefix.length() - 1);
}

std::string getFilename(std::string const &path, bool const ext = true) {
	int extSeparator = -1;
	// Backwards search for directory separator.
	for (int x = path.length(); x > 0; --x) {
		if (path[x] == '/') return ext ? path.substr(x + 1) : path.substr(x + 1, extSeparator - x - 1);
		// If this function was called with "ext" set to false,
		// generate a substring of the path containing the filename
		// only.
		else if (!ext && path[x] == '.' && extSeparator == -1)
			extSeparator = x;
	}
	return path;
}

bool isValidArticle(std::string const &pathPrefix, std::string const &path) {
	return std::regex_search(blog_relative_path(pathPrefix, path), std::regex("^/\\d{4}/\\d{2}/.*\\.md"));
}

std::filesystem::path get_output_path(std::string const &path) {
	std::filesystem::path output = path;
	return output /= "output";
}

void build_dir_structure(std::string const &path) {
	// Create directory tree, which will be used for the website.
	std::filesystem::path rootDir = get_output_path(path);

	// Create output directory if it doesn't exist.
	try {
		if(!std::filesystem::exists(rootDir)) std::filesystem::create_directory(rootDir);
	} catch (std::filesystem::filesystem_error const &e) {
		std::cerr << "error: fs error while generating output website directory structure on \""
			<< path << "\": " << e.what() << std::endl;
		exit(RETURN_FAILED_UNKNOWN_ERROR);
	}

	// Check if a website has already been built.
	//
	// This is a safe measure to prevent overwriting an existing website
	// output, but it can be skipped if explicitly asked by the end-user.
	if (!swg_rt_global_bool("overwrite_existing", false)) {
		std::filesystem::path websiteLock = rootDir /= ".swg_built";
		if (std::filesystem::exists(websiteLock)) {
			std::cout << "error: existing website build detected, exiting...\n";
			exit(RETURN_FAILED_WEBSITE_BUILD_EXISTS);
		}
	}
}

std::string get_template(std::string const &path) {
	// Template lookup
	std::string stPath = path;
	stPath += "/__swg_template.html";
	FILE *swgTemplate = fopen(stPath.c_str(), "r");
	if (swgTemplate == NULL) {
		std::cerr << "error: couldn't open the SWG HTML template file; does it exist?\n";
		perror(stPath.c_str());
		exit(RETURN_FAILED_INVALID_DIRECTORY);
	}

	// Check for content placeholder
	int buflen = 8;
	char buf[buflen];
	std::string htmlTemplate;
	while (fgets(buf, buflen, swgTemplate) != NULL) {
		htmlTemplate += buf;
	}

	fclose(swgTemplate);
	std::cout << "Loaded HTML template into memory.\n";
	return htmlTemplate;
}

void build_blog_structure(std::string const &path, std::string const &prefix, std::list<std::string> const &articles, Blog *blog) {
	std::filesystem::path obp = get_output_path(path); // Output Blog Path
	obp /= "blog";
	obp /= blog->dir;
	try {
		// Create blog directory
		if (!std::filesystem::exists(obp)) std::filesystem::create_directories(obp);
	} catch (std::filesystem::filesystem_error const &e) {
		std::cerr << "error: failed to create directory for blog \"" << blog->name << "\": " << e.what() << std::endl;
		exit(RETURN_FAILED_UNKNOWN_ERROR);
	}

	for (std::string const &a : articles) {
		std::string articlePath = blog_relative_path(prefix, a);
		std::regex yearMonth("(\\d+)");
		auto match = std::sregex_iterator(articlePath.begin(), articlePath.end(), yearMonth);

		std::string year = match->str(),
			month = (++match)->str();

		// TODO: This code could be optimized by removing directory
		// checks for every single article.  Instead, add
		// directory-checks for years and months to a queue (skipping
		// existing ones) and then checking and creating the
		// directories later as needed.
		try {
			// Create directory for the year of this article if it doesn't exist.
			std::filesystem::path oad = obp; // Output Article Directory
			oad /= year;
			if (!std::filesystem::exists(oad)) std::filesystem::create_directory(oad);

			// Do the same for the article month.
			oad /= month;
			if (!std::filesystem::exists(oad)) std::filesystem::create_directory(oad);

			// Now create the article file.
			oad /= getFilename(a, false);
			oad += ".html";
			FILE *articleOutput = fopen(oad.string().c_str(), "w");

			std::string htmlTemplate = get_template(path);
			// NOTE: std::regex requires the C++11 standard.
			std::regex contentPlaceholder("<!--\\[_SWG: \\$CONTENT\\]-->");
			std::string articleHtml = MarkdownParser::make_html(a);
			std::string articleContents = std::regex_replace(htmlTemplate, contentPlaceholder, articleHtml);

			fputs(articleContents.c_str(), articleOutput);
			fclose(articleOutput);
		} catch (std::filesystem::filesystem_error const &e) {
			std::cerr << "error: failed to create directory for an article from blog \""
				<< blog->name << "\": " << e.what() << std::endl;
			exit(RETURN_FAILED_UNKNOWN_ERROR);
		}
	}
}

void build_website(SwgContext &ctx, std::string const &path) {
	build_dir_structure(path);

	// Blog lookup

	bool failed = false;

	for (Blog *b : ctx.blogs) {
		std::list<std::string> parsedArticles = { },
			failedArticles = { };

		// std::filesystem requires the C++17 standard.
		std::string relativePath = path;
		relativePath += "/";
		relativePath += b->path;
		const std::filesystem::path blogPath(relativePath);
		try {
		for (auto const &dir_entry : std::filesystem::recursive_directory_iterator(blogPath)) {
			// Directory item iteration
			std::cout << "\t" << dir_entry << std::endl;

			// Check if the item is a Markdown file.
			std::string filename = getFilename(dir_entry.path());
			if (dir_entry.is_regular_file() &&
					filename.find(".md") == filename.length() - 3) {
				std::cout << "\t\tIs file: " << getFilename(dir_entry.path()) << "\n";
				// Markdown files should be insite a YYYY/MM directory.
				if (!isValidArticle(relativePath, dir_entry.path()))
					failedArticles.insert(failedArticles.end(), dir_entry.path());
				else parsedArticles.insert(parsedArticles.end(), dir_entry.path());
			}
		}
		} catch (std::filesystem::filesystem_error const &e) {
			std::cerr << "error: fs error while attempting to read path for "
				<< b->name << ": " << e.what() << std::endl;
			failed = true;
		}

		/*std::cout << "Parsed " << parsedArticles << " articles;\n" <<
			"Failed to parse " << failedArticles << " articles\n";*/

		for (std::string const &a : parsedArticles)
			std::cout << "Parsed: " << a << "\n";
		for (std::string const &a : failedArticles)
			std::cout << "Unable to parse: " << a << "\n";

		build_blog_structure(path, relativePath, parsedArticles, b);

		free(b);
	}

	if (failed) {
		std::cerr << "Refusing to proceed due to previous errors\n";
		exit(RETURN_FAILED_CONFIG_INVALID_SYNTAX);
	}

	// Website blog

	// Append blog to website
	//ctx.appendBlog(blog1);

	// Generate website
	//ctx.generateWebsite();
}
